# Copyright 2015 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .method import Method


class Event(Method):
    """Scanner for event objects (server-side method)

    Required attributes: `name`

    Optional attributes: `since`

    Child elements: `description` and `arg``
    """
    method_type = 'event'

    def method_args(self):
        """Generator of the arguments to the method

        All arguments to be sent to `._post_event` must be passed in
        """
        for arg in self.args:
            yield arg.name

    def post_args(self):
        """Arguments sent to `._post_event`"""
        for arg in self.args:
            yield arg.name

    def output_doc_param(self, printer):
        """Aguments documented as parameters

        All arguments are event parameters.
        """
        for arg in self.args:
            arg.output_doc_param(printer)

    def output_doc_ret(self, printer):
        """Aguments documented as return values

        Nothing is returned from event calls.
        """
        return

    def output_body(self, printer):
        """Output the body of the event to the printer"""
        args = ', '.join(self.post_args())
        if args:
            printer('self._post_event({:d}, {})'.format(self.opcode, args))
        else:
            printer('self._post_event({:d})'.format(self.opcode))



# Copyright 2015 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .argument import Argument


class Method(object):
    def __init__(self, method, opcode):
        """Scanner for methods

        Corresponds to event and requests defined on an interface
        """
        self._method = method
        self.name = method.attrib['name']

        self.type = method.attrib['type'] if 'type' in method.attrib else None
        self.since = method.attrib['since'] if 'since' in method.attrib else None

        self.opcode = opcode

        # 'global' is a protected name, so append '_'
        if self.name == 'global':
            self.name += '_'

    def scan(self):
        """Scan the method"""
        description = self._method.find('description')
        self.summary = description.attrib['summary']
        if description.text:
            self.description = description.text.strip()
        else:
            self.description = None

        self.args = [Argument(arg) for arg in self._method.findall('arg')]

        for arg in self.args:
            arg.scan()

    def get_imports(self, this_interface):
        """Get the imports required for each of the interfaces"""
        return [arg.get_interface() for arg in self.args
                if arg.interface and arg.interface != this_interface]

    def interface_types(self):
        """Generator of the types (for the wl_interface)"""
        for arg in self.args:
            if arg.interface:
                yield arg.get_interface()
            else:
                if arg.type == 'new_id':
                    yield 'None'
                    yield 'None'
                yield 'None'

    def output(self, printer, in_class):
        """Generate the output for the given method to the printer"""
        signature = ''.join(arg.signature for arg in self.args)
        if self.since and int(self.since) > 1:
            signature = self.since + signature
        args = ', '.join(self.method_args())
        types = ', '.join(self.interface_types())

        printer('@{}.{}("{}", [{}])'.format(
            in_class, self.method_type, signature, types)
        )
        if args:
            printer('def {}(self, {}):'.format(self.name, args))
        else:
            printer('def {}(self):'.format(self.name))
        printer.inc_level()

        # Write the documentation
        self.output_doc(printer)
        # Write out the body of the method
        self.output_body(printer)
        printer.dec_level()

    def output_doc(self, printer):
        """Output the documentation for the interface"""
        if self.description or self.args:
            printer('"""{}'.format(self.summary))
        else:
            # This is a one-line docstring
            printer('"""{}"""'.format(self.summary))
            return

        if self.description:
            printer()
            for line in self.description.split('\n'):
                printer(line.strip())
        # Parameter and returns documentation
        if self.args:
            printer()
            self.output_doc_param(printer)
            self.output_doc_ret(printer)
        printer('"""')



# Copyright 2015 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cffi import FFI

ffi = FFI()

ffi.cdef("""
/******************************************************************************
 * wayland-version.h
 *****************************************************************************/
#define WAYLAND_VERSION_MAJOR ...
#define WAYLAND_VERSION_MINOR ...
#define WAYLAND_VERSION_MICRO ...

/******************************************************************************
 * wayland-util.h
 *****************************************************************************/
// Structs for event and request dispactching
typedef int32_t wl_fixed_t;

struct wl_message {
    const char *name;
    const char *signature;
    const struct wl_interface **types;
};
struct wl_interface {
    const char *name;
    int version;
    int method_count;
    const struct wl_message *methods;
    int event_count;
    const struct wl_message *events;
};
union wl_argument {
    int32_t i; /**< signed integer */
    uint32_t u; /**< unsigned integer */
    wl_fixed_t f; /**< fixed point */
    const char *s; /**< string */
    struct wl_object *o; /**< object */
    uint32_t n; /**< new_id */
    struct wl_array *a; /**< array */
    int32_t h; /**< file descriptor */
};

typedef int (*wl_dispatcher_func_t)(const void *, void *, uint32_t,
                                    const struct wl_message *,
                                    union wl_argument *);

/******************************************************************************
 * wayland-client.h
 *****************************************************************************/
// wl_display methods
struct wl_display* wl_display_connect(const char *name);
struct wl_display* wl_display_connect_to_fd(int fd);
struct wl_event_queue *wl_display_create_queue(struct wl_display *display);
void wl_display_disconnect(struct wl_display *display);
int wl_display_dispatch(struct wl_display *display);
int wl_display_flush(struct wl_display *display);
int wl_display_get_fd(struct wl_display *display);
int wl_display_roundtrip(struct wl_display *display);

// wl_eventqueue methods
void wl_event_queue_destroy(struct wl_event_queue *queue);

// wl_proxy methods
struct wl_proxy *wl_proxy_create(struct wl_proxy *factory,
                                 const struct wl_interface *interface);
void wl_proxy_destroy(struct wl_proxy *proxy);
int wl_proxy_add_dispatcher(struct wl_proxy *proxy,
                            wl_dispatcher_func_t dispatcher_func,
                            const void * dispatcher_data, void *data);
void *wl_proxy_get_user_data(struct wl_proxy *proxy);
void wl_proxy_marshal_array(struct wl_proxy *p, uint32_t opcode,
                            union wl_argument *args);
struct wl_proxy * wl_proxy_marshal_array_constructor(struct wl_proxy *proxy,
                                                     uint32_t opcode, union wl_argument *args,
                                                     const struct wl_interface *interface);
void wl_proxy_set_user_data(struct wl_proxy *proxy, void *user_data);
""")

C = ffi.verify("""
#include <wayland-client.h>
#include <wayland-server.h>
#include <wayland-version.h>
""", libraries=['wayland-client', 'wayland-server'], modulename='_pywayland')



# Copyright 2015 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .entry import Entry


class Enum(object):
    def __init__(self, enum):
        """Scanner for enum objects

        Required attributes: `name` and `since`

        Child elements: `description` and `entry`
        """
        self._enum = enum
        self.name = enum.attrib['name']

        self.since = enum.attrib['since'] if 'since' in enum.attrib else None

    def scan(self):
        """Scan the enum"""
        description = self._enum.find('description')
        if description:
            self.summary = description.attrib['summary']
            if description.text:
                self.description = description.text.strip()
            else:
                self.description = None
        else:
            self.summary = None
            self.description = None

        self.entries = [Entry(entry) for entry in self._enum.findall('entry')]

        for entry in self.entries:
            entry.scan()

    def output(self, printer):
        """Generate the output for the enum to the printer"""
        printer('{0} = enum.Enum("{0}", {{'.format(self.name))
        printer.inc_level()
        for entry in self.entries:
            entry.output(printer)
        printer.dec_level()
        printer('})')



# Copyright 2015 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pywayland.scanner.scanner import Scanner

import argparse
import os

this_dir = os.path.split(__file__)[0]
protocol_dir = os.path.join(this_dir, 'protocol')

xml_file = '/usr/share/wayland/wayland.xml'


def main():
    parser = argparse.ArgumentParser(
        description='Generate wayland protocol files from xml'
    )
    parser.add_argument(
        '-i', '--input', metavar='XML_FILE', default=xml_file, type=str,
        help='Location of input xml file to scan'
    )
    parser.add_argument(
        '-o', '--output', metavar='DIR', default=protocol_dir, type=str,
        help='Location to output protocol files'
    )

    args = parser.parse_args()

    if not os.path.exists(args.output):
        os.makedirs(args.output, 0o775)

    scanner = Scanner(args.input)
    scanner.scan()
    scanner.output(args.output)


if __name__ == '__main__':
    main()



# Copyright 2015 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Entry(object):
    def __init__(self, entry):
        """Scanner for enum entries

        Required attributes: `name` and `value`

        Optional attributes: `summary` and `since`

        Child elements: `description`
        """
        self._entry = entry
        self.name = entry.attrib['name']
        self.value = entry.attrib['value']

        self.summary = entry.attrib['since'] if 'since' in entry.attrib else None
        self.since = entry.attrib['since'] if 'since' in entry.attrib else None

    def scan(self):
        """Scan the enum entry"""
        # Currently no entries implement a description, instead, they have a
        # summary element
        pass

    def output(self, printer):
        """Generate the output for the entry in the enum"""
        printer('"{}": {},'.format(self.name, self.value))



# Copyright 2015 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .interface import Interface
from .printer import Printer

import xml.etree.ElementTree as ET
import os


class Scanner(object):
    def __init__(self, input_file):
        """Main scanner object

        Main scanner object that acts on the input wayland.xml file to generate
        protocol files.

        Required attributes: `name`

        Child elements: `copywright` and `interface`
        """
        self._input_file = os.path.basename(input_file)
        if not os.path.exists(input_file):
            raise ValueError("Input xml file does not exist: {}".format(input_file))
        self._protocol = ET.parse(input_file).getroot()
        if self._protocol.tag != 'protocol':
            raise ValueError("Input file not a valid Wayland protocol file: {}".format(input_file))

        attrib = self._protocol.attrib
        if 'name' not in attrib or attrib['name'] != 'wayland':
            raise ValueError("Input file not a valid Wayland protocol file: {}".format(input_file))

    def __repr__(self):
        return "Scanner({})".format(self._input_file)

    def scan(self):
        """Scan the initialized xml file"""
        self.interfaces = [Interface(iface) for iface in self._protocol.findall('interface')]

        for iface in self.interfaces:
            iface.scan()

    def output(self, output_dir):
        """Output the scanned files to the given directory


        Parameters
        ----------

        output_dir : string
            Path of directory to output protocol files to

        """
        printer = Printer()
        printer.initialize_file()
        path = os.path.join(output_dir, '__init__.py')
        with open(path, 'w') as f:
            printer.write(f)

        for iface in self.interfaces:
            printer = Printer()
            path = os.path.join(output_dir, iface.file_name)

            iface.output(printer)

            with open(path, 'w') as f:
                printer.write(f)



# Copyright 2015 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pywayland import ffi

import re
import weakref

weakkeydict = weakref.WeakKeyDictionary()

re_arg = re.compile(r"(\??)([uifsonah])")


class Message(object):
    def __init__(self, func, signature, types):
        """Wrapper class for wl_message structs

        Base class that correspond to the methods defined on an interface in
        the wayland.xml protocol, and are generated by the scanner.  Subclasses
        specify the type of method, whether it is a server-side or client-side
        method.

        Parameters
        ----------

        func : function
            The function that is represented in the message

        signature : string
            The signature of the arguments of the message

        types : list
            List of the types of any objects included in the argument list,
            None if otherwise.

        """
        self._func = func
        self.name = func.__name__.strip('_')
        self.signature = signature
        self.types = types

        self._ptr = ffi.new('struct wl_message *')
        self._ptr.name = name = ffi.new('char[]', self.name.encode())
        self._ptr.signature = signature = ffi.new('char[]', self.signature.encode())

        self._ptr.types = types = ffi.new('struct wl_interface* []', len(self.types))
        for i, _type in enumerate(self.types):
            if _type:
                self._ptr.types[i] = _type._ptr
            else:
                self._ptr.types[i] = ffi.NULL

        weakkeydict[self._ptr] = (name, signature, types)

    def c_to_arguments(self, args_ptr):
        args = []
        for i, sig_match in enumerate(re_arg.finditer(self.signature)):
            arg_ptr = args_ptr[i]
            null, sig = sig_match.groups()

            # Match numbers (int, unsigned, float, file descriptor)
            if sig == 'i':
                args.append(arg_ptr.i)
            elif sig == 'u':
                args.append(arg_ptr.u)
            elif sig == 'f':
                args.append(arg_ptr.f)
            elif sig == 'h':
                args.append(arg_ptr.h)
            # Match string
            elif sig == 's':
                if arg_ptr == ffi.NULL:
                    if not null:
                        raise Exception
                    args.append(None)
                else:
                    args.append(ffi.string(arg_ptr.s).decode())
            # Object or new id
            elif sig in ('o', 'n'):
                # TODO
                pass
            # Array (i.e. buffer of bytes)
            elif sig == 'a':
                # TODO
                pass

        return args

    def arguments_to_c(self, *args):
        """Create an array of wl_argument C structs

        Generate the CFFI cdata array of wl_argument structs that correspond to
        the arguments of the method.
        """
        nargs = len(re_arg.findall(self.signature))
        args_ptr = ffi.new('union wl_argument []', nargs)

        arg_iter = iter(args)
        refs = []
        for i, sig_match in enumerate(re_arg.finditer(self.signature)):
            null, sig = sig_match.groups()

            # New id (set to null for now, will be assigned on marshal)
            # Then, continue so we don't consume an arg
            if sig == 'n':
                args_ptr[i].o = ffi.NULL
                continue
            arg = next(arg_iter)
            # Match numbers (int, unsigned, float, file descriptor)
            if sig == 'i':
                args_ptr[i].i = arg
            elif sig == 'u':
                args_ptr[i].u = arg
            elif sig == 'f':
                args_ptr[i].f = arg
            elif sig == 'h':
                args_ptr[i].h = arg
            # Match string
            elif sig == 's':
                if arg is None:
                    if not null:
                        raise Exception
                    new_arg = ffi.NULL
                else:
                    new_arg = ffi.new('char []', arg.encode())
                    refs.append(new_arg)
                args_ptr[i].s = new_arg
            # Object
            elif sig == 'o':
                if arg is None:
                    if not null:
                        raise Exception
                    new_arg = ffi.NULL
                else:
                    new_arg = ffi.cast('struct wl_object *', arg._ptr)
                    refs.append(new_arg)
                args_ptr[i].o = new_arg
            # Array (i.e. buffer of bytes)
            elif sig == 'a':
                new_arg = ffi.new('struct wl_array *')
                new_data = ffi.new('void []', len(arg))
                new_arg.alloc = new_arg.size = len(arg)
                ffi.buffer(new_data)[:] = arg
                refs.append(new_arg)
                refs.append(new_data)

        if refs:
            weakkeydict[args_ptr] = refs

        return args_ptr



# Copyright 2015 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .enum import Enum
from .event import Event
from .request import Request

import itertools


class Interface(object):
    def __init__(self, iface):
        """Scanner for interface objects

        Required attributes: `name` and `version`

        Child elements: `description`, `request`, `event`, `enum`
        """
        self._iface = iface
        attrib = iface.attrib
        self.name = attrib['name']
        self.version = int(attrib['version'])

    @property
    def file_name(self):
        """Returns the file name of the interface

        Trim the 'wl_' from the specified interface name.
        """
        return ''.join(self.name.split('_')[1:]) + '.py'

    @property
    def class_name(self):
        """Returns the name of the class of the interface

        Trim the 'wl_' from the specified interface, and capitalize the first
        letter of each word, dropping the underscores.
        """
        return ''.join(x.capitalize() for x in self.name.split('_')[1:])

    def scan(self):
        """Scan the interface"""
        description = self._iface.find('description')
        self.summary = description.attrib['summary']
        self.description = description.text.strip()

        self.enums = [Enum(enum) for enum in self._iface.findall('enum')]
        self.events = [Event(event, i) for i, event in enumerate(self._iface.findall('event'))]
        self.requests = [Request(request, i) for i, request in enumerate(self._iface.findall('request'))]

        for item in itertools.chain(self.enums, self.events, self.requests):
            item.scan()

    def output(self, printer):
        """Generate the output for the interface to the printer"""
        printer.initialize_file()
        printer()

        imports = set(_import for method in itertools.chain(self.requests, self.events)
                      for _import in method.get_imports(self.name))
        printer('from pywayland.interface import Interface, InterfaceMeta')
        for _import in sorted(imports):
            printer('from .{} import {}'.format(_import.lower(), _import))
        printer()
        if self.enums:
            printer('import enum')
        printer('import six')
        printer()
        printer()

        printer('@six.add_metaclass(InterfaceMeta)')
        printer('class {}(Interface):'.format(self.class_name))
        printer.inc_level()
        printer('"""{}'.format(self.summary))
        printer()
        for line in self.description.split('\n'):
            printer(line.strip())
        printer('"""')
        printer('name = "{}"'.format(self.name))
        printer('version = {:d}'.format(self.version))
        for enum in self.enums:
            printer()
            enum.output(printer)
        printer.dec_level()

        for method in itertools.chain(self.requests, self.events):
            printer()
            printer()
            method.output(printer, self.class_name)

        printer()
        printer()
        printer('{}._gen_c()'.format(self.class_name))



# Copyright 2015 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

copyright = """\
# Copyright 2015 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License."""


class Printer(object):
    def __init__(self):
        """Base level printer object

        Allows for storing of lines to be output from the definition of a
        protocol.  Lines are added by directly calling the printer object.
        """
        self.level = 0
        self.lines = []

    def __call__(self, new_line=''):
        """Add the new line to the printer"""
        if new_line:
            self.lines.append(('    ' * self.level) + new_line)
        else:
            self.lines.append('')

    def inc_level(self):
        """Increment the indent level"""
        self.level += 1

    def dec_level(self):
        """Decrement the indent level"""
        self.level -= 1

    def initialize_file(self):
        """Initialize the file by writing out the copywrite"""
        for line in copyright.split('\n'):
            self(line)

    def write(self, _file):
        """Write the lines added to the printer out to the given file"""
        for line in self.lines:
            _file.write(line)
            _file.write('\n')



# Copyright 2015 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .method import Method

# For 'new_id' types with no 'interface'
NO_IFACE = 'interface'
NO_IFACE_VERSION = 'version'


class Request(Method):
    """Scanner for request objects (client-side method)

    Required attributes: `name`

    Optional attributes: `type` and `since`

    Child elements: `description` and `arg`
    """
    method_type = 'request'

    def scan(self):
        """Scan the request object

        In addition to `Method.scan()`, requests need to know if any arguments
        are of type `new_id`.
        """
        # Call the Method.scan() function
        super(Request, self).scan()

        for arg in self.args:
            if arg.type == 'new_id':
                self.new_id = arg
                break
        else:
            self.new_id = None

    def method_args(self):
        """Generator of the arguments to the method

        The `new_id` args are generated in marshaling the args, they do not
        appear in the args of the method.
        """
        for arg in self.args:
            if arg.type == 'new_id':
                if arg.interface:
                    continue
                # A `new_id` with no interface, c.f. wl_registry_bind
                # Need a string (interface name) and int (interface version)
                yield NO_IFACE
                yield NO_IFACE_VERSION
            else:
                yield arg.name

    def marshal_args(self):
        """Arguments sent to `._marshal`"""
        for arg in self.args:
            if arg.type == 'new_id':
                if not arg.interface:
                    yield '{}.name'.format(NO_IFACE)
                    yield NO_IFACE_VERSION
            else:
                yield arg.name

    def output_doc_param(self, printer):
        """Aguments documented as parameters

        Anything that is not a `new_id` is
        """
        for arg in self.args:
            if arg.type == 'new_id':
                if arg.interface:
                    continue
                printer(':param {}: Interface name'.format(NO_IFACE))
                printer(':type {}: `string`'.format(NO_IFACE))
                printer(':param {}: Interface version'.format(NO_IFACE_VERSION))
                printer(':type {}: `int`'.format(NO_IFACE_VERSION))
            else:
                arg.output_doc_param(printer)

    def output_doc_ret(self, printer):
        """Aguments documented as return values

        Arguments of type `new_id` are returned from requests.
        """
        for arg in self.args:
            if arg.type == 'new_id':
                arg.output_doc_ret(printer)

    def output_body(self, printer):
        """Output the body of the request to the printer"""
        args = ', '.join(self.marshal_args())
        if self.new_id:
            if self.new_id.interface:
                id_class = self.new_id.get_interface()
            else:
                id_class = NO_IFACE
            id_name = self.new_id.name
            if args:
                printer('{} = self._marshal_constructor({:d}, {}, {})'.format(
                    id_name, self.opcode, id_class, args)
                )
            else:
                printer('{} = self._marshal_constructor({:d}, {})'.format(
                    id_name, self.opcode, id_class)
                )
            printer('return {}'.format(id_name))
        else:
            if args:
                printer('self._marshal({:d}, {})'.format(self.opcode, args))
            else:
                printer('self._marshal({:d})'.format(self.opcode))
            if self.type == 'destructor':
                printer('self._destroy()')



# Copyright 2015 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pywayland import ffi, C
from pywayland.protocol.display import Display as _Display

from .proxy import Proxy, _dispatcher


class Display(Proxy):
    _interface = _Display
    _ptr = None

    def __init__(self):
        """Represents a connection to the compositor

        A wl_display object represents a client connection to a Wayland
        compositor. It is created with either wl_display_connect() or
        wl_display_connect_to_fd(). A connection is terminated using
        wl_display_disconnect().

        A wl_display is also used as the wl_proxy for the wl_display singleton
        object on the compositor side.

        A wl_display object handles all the data sent from and to the
        compositor. When a wl_proxy marshals a request, it will write its wire
        representation to the display's write buffer. The data is sent to the
        compositor when the client calls wl_display_flush().

        Incoming data is handled in two steps: queueing and dispatching. In the
        queue step, the data coming from the display fd is interpreted and
        added to a queue. On the dispatch step, the handler for the incoming
        event set by the client on the corresponding wl_proxy is called.

        A wl_display has at least one event queue, called the _default queue_.
        Clients can create additional event queues with
        wl_display_create_queue() and assign wl_proxy's to it. Events occurring
        in a particular proxy are always queued in its assigned queue.  A
        client can ensure that a certain assumption, such as holding a lock or
        running from a given thread, is true when a proxy event handler is
        called by assigning that proxy to an event queue and making sure that
        this queue is only dispatched when the assumption holds.

        The default queue is dispatched by calling wl_display_dispatch().  This
        will dispatch any events queued on the default queue and attempt to
        read from the display fd if it's empty. Events read are then queued on
        the appropriate queues according to the proxy assignment.

        A user created queue is dispatched with wl_display_dispatch_queue().
        This function behaves exactly the same as wl_display_dispatch() but it
        dispatches given queue instead of the default queue.

        A real world example of event queue usage is Mesa's implementation of
        glSwapBuffers() for the Wayland platform. This function might need o
        block until a frame callback is received, but dispatching the default
        ueue could cause an event handler on the client to start drawing gain.
        This problem is solved using another event queue, so that only he
        events handled by the EGL code are dispatched during the block.

        This creates a problem where a thread dispatches a non-default queue,
        reading all the data from the display fd. If the application would call
        _poll(2)_ after that it would block, even though there might be events
        queued on the default queue. Those events should be dispatched with
        wl_display_dispatch_(queue_)pending() before flushing and blocking.
        """
        super().__init__()

    def __del__(self):
        self.disconnect()

    def connect(self, name_or_fd=None):
        """Connect to a Wayland display

        Connect to the Wayland display by name of fd.  An `int` parameter opens
        the connection using the file descriptor.  The wl_display takes
        ownership of the fd and will close it when the display is destroyed.
        The fd will also be closed in case of failure.  A string will open the
        display of the given name.  If name is None, its value will be replaced
        with the WAYLAND_DISPLAY environment variable if it is set, otherwise
        display "wayland-0" will be used.
        """
        if isinstance(name_or_fd, int):
            self._ptr = C.wl_display_connect_to_fd(name_or_fd)
        else:
            if name_or_fd is None:
                name = ffi.NULL
            else:
                name = name_or_fd.encode()
            self._ptr = C.wl_display_connect(name)

        if self._ptr == ffi.NULL:
            raise Exception

        _ptr = ffi.cast('struct wl_proxy *', self._ptr)
        C.wl_proxy_add_dispatcher(_ptr, _dispatcher, self._handle, ffi.NULL)

    def disconnect(self):
        """Close a connection to a Wayland display

        Close the connection to display and free all resources associated with
        it.
        """
        if self._ptr:
            C.wl_display_disconnect(self._ptr)
            self._ptr = None

    def get_fd(self):
        """Get a display context's file descriptor

        Return the file descriptor associated with a display so it can be
        integrated into the client's main loop.
        """
        return C.wl_display_get_fd(self._ptr)

    def dispatch(self):
        """Process incoming events

        Dispatch the display's default event queue.

        If the default event queue is empty, this function blocks until there
        are events to be read from the display fd. Events are read and queued
        on the appropriate event queues. Finally, events on the default event
        queue are dispatched.

        In multi-threaded environment, programmer may want to use
        wl_display_read_events(). However, use of wl_display_read_events() must
        not be mixed with wl_display_dispatch(). See wl_display_read_events()
        and wl_display_prepare_read() for more details.

        Note: It is not possible to check if there are events on the queue or
        not. For dispatching default queue events without blocking, see
        wl_display_dispatch_pending().
        """
        return C.wl_display_dispatch(self._ptr)

    def flush(self):
        """Send all buffered requests on the display to the server

        Send all buffered data on the client side to the server. Clients
        should call this function before blocking. On success, the number
        of bytes sent to the server is returned. On failure, this
        function returns -1 and errno is set appropriately.

        wl_display_flush() never blocks.  It will write as much data as
        possible, but if all data could not be written, errno will be set
        to EAGAIN and -1 returned.  In that case, use poll on the display
        file descriptor to wait for it to become writable again.
        """
        return C.wl_display_flush(self._ptr)

    def roundtrip(self):
        """Block until all pending request are processed by the server

        Blocks until the server process all currently issued requests and
        sends out pending events on the default event queue.
        """
        return C.wl_display_roundtrip(self._ptr)

for msg in _Display.requests:
    setattr(Display, msg.name, msg._func)



# Copyright 2015 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pywayland.client.proxy import ProxyMeta
from pywayland import ffi

from .message import Message

import weakref

weakkeydict = weakref.WeakKeyDictionary()


class Interface(object):
    """Wrapper class for wl_wayland structs

    Base class for interfaces that are defined by the wayland.xml class and
    generated by the scanner.  Sub-classes should use the InterfaceMeta
    metaclass, which will define subcls.events and subcls.requests, the lists
    of the methods on this interface.  These class variables are populated by
    the event and request decorators.
    """

    @classmethod
    def event(cls, signature, types):
        """Decorator for interface events

        Adds the decorated method to the list of events of the interface
        (server-side method).

        Parameters
        ----------

        signature : string
            Encodes the types of the arguments to the decorated function.

        types : list
            List of the types of any objects included in the argument list,
            None if otherwise.

        """
        def wrapper(func):
            cls.events.append(Message(func, signature, types))
            return func
        return wrapper

    @classmethod
    def request(cls, signature, types):
        """Decorator for interface requests

        Adds the decorated method to the list of requests of the interface
        (client-side method).

        Parameters
        ----------

        signature : string
            Encodes the types of the arguments to the decorated function.

        types : list
            List of the types of any objects included in the argument list,
            None if otherwise.

        """
        def wrapper(func):
            cls.requests.append(Message(func, signature, types))
            return func
        return wrapper

    @classmethod
    def proxy_class(interface, proxy_ptr):
        """Return a proxy class for the given interface and wl_proxy pointer"""
        new_class = ProxyMeta(
            interface.name, (), {'_interface': interface, '_ptr': proxy_ptr}
        )
        return new_class()

    @classmethod
    def _gen_c(cls):
        """Creates the wl_interface C struct

        Generates the CFFI cdata for the wl_interface struct given by the
        interface.

        """
        # Initialize the interface cdata
        cls._ptr.name = name = ffi.new('char[]', cls.name.encode())
        cls._ptr.version = cls.version

        # Determine the number of methods to assign and assign them
        cls._ptr.method_count = len(cls.requests)
        cls._ptr.methods = methods_ptr = ffi.new("struct wl_message[]", len(cls.requests))
        # Iterate over the methods
        for i, message in enumerate(cls.requests):
            # First, generate the wl_message cdata
            msg_buf = ffi.buffer(message._ptr)
            methods_buf = ffi.buffer(methods_ptr + i)
            # Copy the contents of the cdata into the allocated cdata
            methods_buf[:] = msg_buf

        cls._ptr.event_count = len(cls.events)
        cls._ptr.events = events_ptr = ffi.new("struct wl_message[]", len(cls.events))
        # Iterate over the methods
        for i, message in enumerate(cls.events):
            # First, generate the wl_message cdata
            msg_buf = ffi.buffer(message._ptr)
            events_buf = ffi.buffer(events_ptr + i)
            # Copy the contents of the cdata into the allocated cdata
            events_buf[:] = msg_buf

        weakkeydict[cls._ptr] = (name, methods_ptr, events_ptr)


class InterfaceMeta(type):
    def __init__(self, name, bases, dct):
        """Metaclass for Interfaces

        Initializes empty lists for events and requests for the given class.
        Also allocates a cdata struct for the Interface.
        """
        self._ptr = ffi.new("struct wl_interface *")
        self.events = []
        self.requests = []



# Copyright 2015 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import os
import platform
import sys

from setuptools import setup
from distutils.command.build import build

protocol_dir = './pywayland/protocol'


def build_protocol(input_file):
    """Build the ./pywayland/protocol/ directory from the given xml file"""
    from pywayland.scanner.scanner import Scanner

    # Ensure the protocol dir exists
    if not os.path.isdir(protocol_dir):
        os.makedirs(protocol_dir, 0o775)

    # Run and scan the xml file
    scanner = Scanner(input_file)
    scanner.scan()
    scanner.output(protocol_dir)


# Adapted from http://github.com/xattr/xattr
class module_build(build):
    """
    This is a shameful hack to ensure that cffi is present when we specify
    ext_modules. We can't do this eagerly because setup_require hasn't run yet.

    Furthermore, we also need to build the pywayland.protocol module and add it
    to the packages.  This must also be held off until setup_require has run.
    """
    user_options = build.user_options + [
        ('xml-file=', None, 'Location of wayland.xml protocol file')
    ]

    def initialize_options(self):
        build.initialize_options(self)
        self.xml_file = '/usr/share/wayland/wayland.xml'

    def finalize_options(self):
        from pywayland import ffi, __wayland_version__
        self.distribution.ext_modules = [ffi.verifier.get_extension()]
        build.finalize_options(self)

    def run(self):
        build_protocol(self.xml_file)
        # Now we can add the protocols to the packages
        self.distribution.packages.append('pywayland.protocol')

        build.run(self)

        self.distribution.long_description = self.distribution.long_description.format(
            __wayland_version__
        )



description = 'Python bindings for the libwayland library written in pure Python'

try:
    rst_input = open('README.rst').read().split('\n')
except:
    long_description = ""
else:
    rst_head = rst_input[:3]
    rst_body = rst_input[3:]

    long_description = '\n'.join(
        rst_head + ['Built against Wayland {}', ''] + rst_body
    )

classifiers = [
    'Development Status :: 2 - Pre-Alpha',
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: POSIX',
    'Operating System :: POSIX :: Linux',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.2',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: Implementation :: CPython',
    'Programming Language :: Python :: Implementation :: PyPy',
    'Topic :: Desktop Environment :: Window Managers',
    'Topic :: Software Development :: Libraries'
]

dependencies = ['six>=1.4.1']

modules = [
    'pywayland.client',
    'pywayland.scanner'
]

if sys.version_info < (3, 4):
    dependencies.append('enum34')

if platform.python_implementation() != "PyPy":
    dependencies.append('cffi>=0.9')

setup(
    name='pywayland',
    version='0.0.1a.dev1',
    author='Sean Vig',
    author_email='sean.v.775@gmail.com',
    url='http://github.com/flacjacket/pywayland',
    description=description,
    license='Apache License 2.0',
    classifiers=classifiers,
    install_requires=dependencies,
    setup_requires=dependencies,
    packages=['pywayland'] + modules,
    entry_points={
        'console_scripts': [
            'pywayland-scanner = pywayland.pywayland_scanner:main'
        ]
    },
    zip_safe=False,
    ext_package='_pywayland',
    cmdclass={
        'build': module_build
    }
)



# Copyright 2015 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pywayland import ffi, C
import re

re_args = re.compile(r'(\??)([fsoanuih])')


@ffi.callback("int(void *, void *, uint32_t, struct wl_message *, union wl_argument *)")
def _dispatcher(data, target, opcode, message, c_args):
    # `data` is the handle to self
    # `target` is the wl_proxy for self
    # `message` is the wl_message for self._interface.events[opcode]
    # TODO: handle any user_data attached to the wl_proxy
    self = ffi.from_handle(data)
    args = self._interface.events[opcode].c_to_arguments(c_args)

    if opcode in self.listener:
        self.listener[opcode](self, *args)

    return 0


class Proxy(object):
    def __init__(self):
        """Represents a protocol object on the client side.

        A `Proxy` acts as a client side proxy to an object existing in the
        compositor. Events coming from the compositor are also handled by the
        proxy, which will in turn call the handler set with
        `Proxy.add_listener()`.
        """
        self.listener = {}
        self.user_data = None
        self._user_data = None

        self._handle = _handle = ffi.new_handle(self)

        # This should only be true for wl_display proxies, as they will
        # initialize its pointer on a `.connect()` call
        if self._ptr is None:
            return

        _ptr = ffi.cast('struct wl_proxy *', self._ptr)
        C.wl_proxy_add_dispatcher(_ptr, _dispatcher, _handle, ffi.NULL)

    def add_listener(self, callback, opcode=None, name=None):
        """Add listener for an event

        Add listener for an event, either by opcode (passed as an int) or by
        the name of the event, `name` (passed as a string).  Exactly one of
        these parameters must be set.
        """
        if (opcode and name) or not (opcode or name):
            raise ValueError("Exactly one of `opcode` or `name` can be set")

        if name:
            for opcode, event in self._interface.events.items():
                if event.name == name:
                    break
            else:
                raise ValueError("Could not find event with name {}".format(name))

        self.listener[opcode] = callback

    def get_listener(self):
        """Get the listeners for the interface events"""
        return self.listener

    def set_user_data(self, data):
        """Set the user data for the proxy"""
        self.user_data = data
        _ptr = ffi.cast('struct wl_proxy *', self._ptr)
        C.wl_proxy_set_user_data(_ptr, data)

    def get_user_data(self):
        """Get the user data for the proxy"""
        return self.user_data

    def _destroy(self):
        if self._ptr:
            _ptr = ffi.cast('struct wl_proxy *', self._ptr)
            C.wl_proxy_destroy(_ptr)
            self._ptr = None

    def _marshal(self, opcode, *args):
        # Create a wl_argument array
        args_ptr = self._interface.requests[opcode].arguments_to_c(*args)
        # Make the cast to a wl_proxy
        proxy = ffi.cast('struct wl_proxy *', self._ptr)

        C.wl_proxy_marshal_array(proxy, opcode, args_ptr)

    def _marshal_constructor(self, opcode, interface, *args):
        # Create a wl_argument array
        args_ptr = self._interface.requests[opcode].arguments_to_c(*args)
        # Make the cast to a wl_proxy
        proxy = ffi.cast('struct wl_proxy *', self._ptr)

        proxy_ptr = C.wl_proxy_marshal_array_constructor(
            proxy, opcode, args_ptr, interface._ptr
        )
        return interface.proxy_class(proxy_ptr)


class ProxyMeta(type):
    """Metaclass to initialize proxy classes from interfaces

    The base class must define an `._interface` parameter corresponding to a
    `Interface`, a `._ptr` parameter corresponding to the `wl_interface`
    pointer, and a `.requests` parameter giving a list of `Message`'s.  This
    casts the `_ptr` to the correct type and adds the functions of each method
    to the class as methods.
    """
    def __new__(cls, name, bases, dct):
        assert '_interface' in dct
        assert '_ptr' in dct

        # We'll construct the proxy class from the interface
        interface = dct['_interface']

        # Cast the pointer to the right type
        dct['_ptr'] = ffi.cast('struct {} *'.format(interface.name), dct['_ptr'])

        # Get name from interface
        new_name = ''.join(x.capitalize() for x in interface.name.split('_')[1:])

        # Create new class
        new_class = type(new_name, (Proxy, ), dct)
        # Add requests out of interface to new class
        for msg in interface.requests:
            setattr(new_class, msg.name, msg._func)

        return new_class



# Copyright 2015 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

NO_IFACE_NAME = 'interface'


class Argument(object):
    def __init__(self, arg):
        """Argument to a request or event method

        Required attributes: `name` and `type`

        Optional attributes: `summary`, `interface`, and `allow-null`

        Child elements: `description`
        """
        self._arg = arg
        self.name = arg.attrib['name']
        self.type = arg.attrib['type']
        self.summary = arg.attrib['summary'] if 'summary' in arg.attrib else None
        self.interface = arg.attrib['interface'] if 'interface' in arg.attrib else None
        self.allow_null = arg.attrib['allow-null'] == 'true' if 'allow-null' in arg.attrib else False

    def scan(self):
        """Scan the argument"""
        # Currently, no arguments have description elements, but the dtd says
        # they may, they should be parsed here
        pass

    def get_interface(self):
        """Returns the Interface class name

        Gives the class name for the Interface coresponding to the type of the
        argument.
        """
        return ''.join(x.capitalize() for x in self.interface.split('_')[1:])

    @property
    def signature(self):
        """Return the signature of the argument

        Return the string corresponding to the signature of the argument as it
        appears in the signature of the wl_message struct.
        """
        null_string = '?' if self.allow_null else ''
        return null_string + self.type_to_string()

    def type_to_string(self):
        """Translate type to signature string"""
        if self.type == 'int':
            return 'i'
        elif self.type == 'uint':
            return 'u'
        elif self.type == 'fixed':
            return 'f'
        elif self.type == 'string':
            return 's'
        elif self.type == 'object':
            return 'o'
        elif self.type == 'new_id':
            if self.interface:
                return 'n'
            else:
                return 'sun'
        elif self.type == 'array':
            return 'a'
        elif self.type == 'fd':
            return 'h'

    def output_doc_param(self, printer):
        """Document the argument as a parameter"""
        # Output the parameter and summary
        if self.summary:
            printer(':param {}: {}'.format(self.name, self.summary))
        else:
            printer(':param {}:'.format(self.name))

        # Determine the type to be output
        if self.interface:
            arg_type = ':class:`{}`'.format(self.get_interface())
        else:
            arg_type = '`{}`'.format(self.type)

        # Output the parameter type
        if self.allow_null:
            printer(':type {}: {} or `None`'.format(self.name, arg_type))
        else:
            printer(':type {}: {}'.format(self.name, arg_type))

    def output_doc_ret(self, printer):
        """Document the argument as a return"""
        # Determine the type to be output
        if self.interface:
            arg_type = ':class:`{}`'.format(self.get_interface())
        else:
            # Only new_id's are returned, the only corner case here is for
            # wl_registry.bind, so no interface => Proxy
            arg_type = ':class:`Proxy` of specified Interface'

        # Output the type and summary
        if self.summary:
            printer(':returns: {} -- {}'.format(arg_type, self.summary))
        else:
            printer(':returns: {}'.format(arg_type))
